#!/usr/bin/env python
# Copyright (c) 2013-2019 Andrea Bonomi <andrea.bonomi@gmail.com>
#
# Published under the terms of the MIT license.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
#

__author__ = "Andrea Bonomi <andrea.bonomi@gmail.com>"
__license__ = "MIT"
__version__ = "5.3"
__date__ = "15 August 2013"

from typing import Any, Dict, Optional, Type, Union

from .abstract import AbstractCEnum, AbstractCStruct, CStructMeta
from .base import (
    BIG_ENDIAN,
    CHAR_ZERO,
    DEFINES,
    ENUMS,
    LITTLE_ENDIAN,
    NATIVE_ORDER,
    STRUCTS,
    TYPEDEFS,
)
from .c_parser import parse_struct_def
from .cenum import CEnum
from .cstruct import CStruct
from .mem_cstruct import MemCStruct
from .native_types import get_native_type

__all__ = [
    "LITTLE_ENDIAN",
    "BIG_ENDIAN",
    "NATIVE_ORDER",
    "CHAR_ZERO",
    "CStruct",
    "MemCStruct",
    "CEnum",
    "define",
    "undef",
    "getdef",
    "typedef",
    "get_type",
    "sizeof",
    "parse",
]


def define(key: str, value: Any) -> None:
    """
    Define a constant that can be used in the C struct

    Examples:
        >>> define("INIT_THREAD_SIZE", 16384)

    Args:
        key: identifier
        value: value of the constant
    """
    DEFINES[key] = value


def undef(key: str) -> None:
    """
    Undefine a symbol that was previously defined with define

    Examples:
        >>> define("INIT_THREAD_SIZE", 16384)
        >>> undef("INIT_THREAD_SIZE")

    Args:
        key: identifier

    Raises:
        KeyError: If key is not defined
    """
    del DEFINES[key]


def getdef(key: str) -> Any:
    """
    Return the value for a constant

    Examples:
        >>> define("INIT_THREAD_SIZE", 16384)
        >>> getdef("INIT_THREAD_SIZE")

    Args:
        key: identifier

    Raises:
        KeyError: If key is not defined
    """
    return DEFINES[key]


def typedef(type_: str, alias: str) -> None:
    """
    Define an alias name for a data type

    Examples:
        >>> typedef("int", "status")
        >>> sizeof("status")
        4

    Args:
        type_: data type
        alias: new alias name
    """
    TYPEDEFS[alias] = type_


def get_type(type_: str) -> Any:
    """
    Get a data type (struct, union, enum) by name

    Examples:
        >>> get_type("struct Position")
        <class 'abc.Position'>
        >>> get_type("enum htmlfont")
        <enum 'htmlfont'>

    Args:
        type_: C type, struct or union (e.g. 'short int' or 'struct ZYZ'), enum or native type

    Returns:
        class: data type class

    Raises:
        KeyError: If type is not defined
    """
    while type_ in TYPEDEFS:
        type_ = TYPEDEFS[type_]
    if isinstance(type_, CStructMeta):
        return type_
    elif type_.startswith("struct ") or type_.startswith("union "):
        kind, type_ = type_.split(" ", 1)
        try:
            return STRUCTS[type_]
        except KeyError:
            raise KeyError(f"Unknown {kind} `{type_}`")
    elif type_.startswith("enum "):
        kind, type_ = type_.split(" ", 1)
        try:
            return ENUMS[type_]
        except KeyError:
            raise KeyError(f"Unknown {kind} `{type_}`")
    else:
        return get_native_type(type_)


def sizeof(type_: str) -> int:
    """
    Return the size of the type.

    Examples:
        >>> sizeof("struct Position")
        16
        >>> sizeof('enum htmlfont')
        4
        >>> sizeof("int")
        4

    Args:
        type_: C type, struct or union (e.g. 'short int' or 'struct ZYZ'), enum or native type

    Returns:
        size: size in bytes

    Raises:
        KeyError: If type is not defined
    """
    while type_ in TYPEDEFS:
        type_ = TYPEDEFS[type_]
    data_type = get_type(type_)
    return data_type.sizeof()


def parse(
    __struct__: str, __cls__: Optional[Type[AbstractCStruct]] = None, **kargs: Dict[str, Any]
) -> Union[Type[AbstractCStruct], Type[AbstractCEnum], None]:
    """
    Return a new class mapping a C struct/union/enum definition.
    If the string does not contains any definition, return None.
    If the string contains multiple struct/union/enum definitions, returns the last definition.

    Examples:
        >>> cstruct.parse('struct Pair { unsigned char a; unsigned char b; };')
        <class 'abc.Pair'>

    Args:
        __struct__ (str): definition of the struct (or union/enum) in C syntax
        __cls__ (type): super class - CStruct(default) or MemCStruct
        __byte_order__ (str): byte order, valid values are LITTLE_ENDIAN, BIG_ENDIAN, NATIVE_ORDER

    Returns:
        cls: __cls__ subclass

    Raises:
        cstruct.exceptions.ParserError: Parsing exception

    """
    if __cls__ is None:
        __cls__ = MemCStruct
    cls_def = parse_struct_def(__struct__, __cls__=__cls__, process_muliple_definition=True, **kargs)
    if cls_def is None:
        return None
    return cls_def["__cls__"].parse(cls_def, **kargs)



#!/usr/bin/env python
# Copyright (c) 2013-2019 Andrea Bonomi <andrea.bonomi@gmail.com>
#
# Published under the terms of the MIT license.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
#

import struct
from abc import ABCMeta
from typing import Any, Dict, Tuple, Type

__all__ = [
    "get_native_type",
    "AbstractNativeType",
    "Char",
    "SignedChar",
    "UnsignedChar",
    "Short",
    "UnsignedShort",
    "Int",
    "UnsignedInt",
    "Long",
    "UnsignedLong",
    "LongLong",
    "UnsignedLongLong",
    "Float",
    "Double",
    "Pointer",
    "Int8",
    "UnsignedInt8",
    "Int16",
    "UnsignedInt16",
    "Int32",
    "UnsignedInt32",
    "Int64",
    "UnsignedInt64",
]


NATIVE_TYPES: Dict[str, "AbstractNativeType"] = {}


def get_native_type(type_: str) -> "AbstractNativeType":
    """
    Get a base data type by name

    Args:
        type_: data type

    Returns:
        class: data type class

    Raises:
        KeyError: If type is not defined
    """
    try:
        return NATIVE_TYPES[type_]
    except KeyError:
        raise KeyError(f"Unknown type `{type_}`")


class NativeTypeMeta(ABCMeta):
    __size__: int = 0
    " Size in bytes "
    type_name: str = ""
    " Type name "
    native_format: str = ""
    " Type format "

    def __new__(metacls: Type[type], name: str, bases: Tuple[str], namespace: Dict[str, Any]) -> Type[Any]:
        if namespace.get("native_format"):
            native_format = namespace["native_format"]
            namespace["__size__"] = struct.calcsize(native_format)
        else:
            native_format = None
            namespace["native_format"] = None
            namespace["__size__"] = None
        new_class: Type[Any] = super().__new__(metacls, name, bases, namespace)
        if namespace.get("type_name"):
            NATIVE_TYPES[namespace["type_name"]] = new_class
        return new_class

    def __len__(cls) -> int:
        "Type size (in bytes)"
        return cls.__size__

    @property
    def size(cls) -> int:
        "Type size (in bytes)"
        return cls.__size__


class AbstractNativeType(metaclass=NativeTypeMeta):
    def __str__(self) -> str:
        return self.type_name

    @classmethod
    def sizeof(cls) -> int:
        "Type size (in bytes)"
        return cls.__size__


class Char(AbstractNativeType):
    type_name = "char"
    native_format = "s"


class SignedChar(AbstractNativeType):
    type_name = "signed char"
    native_format = "b"


class UnsignedChar(AbstractNativeType):
    type_name = "unsigned char"
    native_format = "B"


class Short(AbstractNativeType):
    type_name = "short"
    native_format = "h"


class UnsignedShort(AbstractNativeType):
    type_name = "unsigned short"
    native_format = "H"


class Int(AbstractNativeType):
    type_name = "int"
    native_format = "i"


class UnsignedInt(AbstractNativeType):
    type_name = "unsigned int"
    native_format = "I"


class Long(AbstractNativeType):
    type_name = "long"
    native_format = "l"


class UnsignedLong(AbstractNativeType):
    type_name = "unsigned long"
    native_format = "L"


class LongLong(AbstractNativeType):
    type_name = "long long"
    native_format = "q"


class UnsignedLongLong(AbstractNativeType):
    type_name = "unsigned long long"
    native_format = "Q"


class Float(AbstractNativeType):
    type_name = "float"
    native_format = "f"


class Double(AbstractNativeType):
    type_name = "double"
    native_format = "d"


class Pointer(AbstractNativeType):
    type_name = "void *"
    native_format = "P"


class Int8(AbstractNativeType):
    type_name = "int8"
    native_format = "b"


class UnsignedInt8(AbstractNativeType):
    type_name = "uint8"
    native_format = "B"


class Int16(AbstractNativeType):
    type_name = "int16"
    native_format = "h"


class UnsignedInt16(AbstractNativeType):
    type_name = "uint16"
    native_format = "H"


class Int32(AbstractNativeType):
    type_name = "int32"
    native_format = "i"


class UnsignedInt32(AbstractNativeType):
    type_name = "uint32"
    native_format = "I"


class Int64(AbstractNativeType):
    type_name = "int64"
    native_format = "q"


class UnsignedInt64(AbstractNativeType):
    type_name = "uint64"
    native_format = "Q"



#!/usr/bin/env python
# Copyright (c) 2013-2019 Andrea Bonomi <andrea.bonomi@gmail.com>
#
# Published under the terms of the MIT license.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
#

import copy
import struct
from enum import Enum
from typing import TYPE_CHECKING, Any, List, Optional, Type

from .base import NATIVE_ORDER
from .exceptions import ParserError
from .native_types import get_native_type

if TYPE_CHECKING:
    from .abstract import AbstractCStruct

__all__ = ["align", "calculate_padding", "Kind", "FieldType"]


def align(byte_order: Optional[str]) -> bool:
    return byte_order is None or byte_order == NATIVE_ORDER


def calculate_padding(byte_order: Optional[str], alignment: int, pos: int) -> int:
    if align(byte_order):  # calculate the padding
        modulo = pos % alignment
        if modulo:  # not aligned
            return alignment - modulo
    return 0


class Kind(Enum):
    """
    Field type
    """

    NATIVE = 0
    "Native type (e.g. int, char)"
    STRUCT = 1
    "Struct type"
    UNION = 2
    "Union type"
    ENUM = 3
    "Enum type"


class FieldType(object):
    """
    Struct/Union field

    Attributes:
        kind (Kind): struct/union/native
        c_type (str): field type
        ref (AbstractCStruct): struct/union class ref
        vlen (int): number of elements
        flexible_array (bool): True for flexible arrays
        offset (int): relative memory position of the field (relative to the struct)
        padding (int): padding
    """

    def __init__(
        self,
        kind: Kind,
        c_type: str,
        ref: Optional[Type["AbstractCStruct"]],
        vlen: int,
        flexible_array: bool,
        byte_order: Optional[str],
        offset: int,
    ) -> None:
        """
        Initialize a Struct/Union field

        Args:
            kind: struct/union/native
            c_type: field type
            ref: struct/union class ref
            vlen: number of elements
            flexible_array: True for flexible arrays
            offset: relative memory position of the field (relative to the struct)
        """
        self.kind = kind
        self.c_type = c_type
        self.ref = ref
        self.vlen = vlen
        self.flexible_array = flexible_array
        self.byte_order = byte_order
        self.offset = self.base_offset = offset
        self.padding = 0

    def unpack_from(self, buffer: bytes, offset: int = 0) -> Any:
        """
        Unpack bytes containing packed C structure data

        Args:
            buffer: bytes to be unpacked
            offset: optional buffer offset

        Returns:
            data: The unpacked data
        """
        if self.is_native or self.is_enum:
            result = struct.unpack_from(self.fmt, buffer, self.offset + offset)

            if self.is_enum:
                result = tuple(map(self.ref, result))

            if self.is_array:
                return list(result)
            else:
                return result[0]
        else:  # struct/union
            if self.vlen == 1:  # single struct/union
                instance: AbstractCStruct = self.ref()  # type: ignore
                instance.unpack_from(buffer, self.offset + offset)
                return instance
            else:  # multiple struct/union
                instances: List[AbstractCStruct] = []
                for j in range(0, self.vlen):
                    instance: AbstractCStruct = self.ref()  # type: ignore
                    instance.unpack_from(buffer, self.offset + offset + j * instance.size)
                    instances.append(instance)
                return instances

    def pack(self, data: Any) -> bytes:
        """
        Pack the field into bytes

        Args:
            data: data to be packed

        Returns:
            bytes: The packed structure
        """
        if self.flexible_array:
            self.vlen = len(data)  # set flexible array size
            return struct.pack(self.fmt, *data)
        elif self.is_array:
            return struct.pack(self.fmt, *data)
        else:
            return struct.pack(self.fmt, data)

    @property
    def is_array(self) -> bool:
        "True if field is an array/flexible array"
        return self.flexible_array or (not (self.vlen == 1 or self.c_type == "char"))

    @property
    def is_native(self) -> bool:
        "True if the field is a native type (e.g. int, char)"
        return self.kind == Kind.NATIVE

    @property
    def is_enum(self) -> bool:
        "True if the field is an enum"
        return self.kind == Kind.ENUM

    @property
    def is_struct(self) -> bool:
        "True if the field is a struct"
        return self.kind == Kind.STRUCT

    @property
    def is_union(self) -> bool:
        "True if the field is an union"
        return self.kind == Kind.UNION

    @property
    def native_format(self) -> str:
        "Field format (struct library format)"
        if self.is_native:
            try:
                return get_native_type(self.c_type).native_format
            except KeyError:
                raise ParserError(f"Unknow type `{self.c_type}`")
        elif self.is_enum:
            return self.ref.__native_format__
        else:
            return "c"

    @property
    def fmt(self) -> str:
        "Field format prefixed by byte order (struct library format)"
        if self.is_native or self.is_enum:
            fmt = (str(self.vlen) if self.vlen > 1 or self.flexible_array else "") + self.native_format
        else:  # Struct/Union
            fmt = str(self.vlen * self.ref.sizeof()) + self.native_format
        if self.byte_order:
            return self.byte_order + fmt
        else:
            return fmt

    @property
    def vsize(self) -> int:
        "Field size in bytes"
        return struct.calcsize(self.fmt)

    @property
    def alignment(self) -> int:
        "Alignment"
        if self.is_native or self.is_enum:
            if self.byte_order is not None:
                return struct.calcsize(self.byte_order + self.native_format)
            else:
                return struct.calcsize(self.native_format)
        else:  # struct/union
            return self.ref.__alignment__

    def align_filed_offset(self) -> None:
        "If the byte order is native, align the field"
        if align(self.byte_order) and self.c_type != "char":
            self.padding = calculate_padding(self.byte_order, self.alignment, self.base_offset)
            self.offset = self.base_offset + self.padding

    def copy(self) -> "FieldType":
        "Return a shallow copy of this FieldType"
        return copy.copy(self)

    def __repr__(self) -> str:  # pragma: no cover
        return repr(self.__dict__)
